/*
 * Indie Game Jame 2 - example game "framework"
 * 2004/3/15 checker@d6.com
 */
#include "shared/core/first.h"

// windows crap
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <windowsx.h>

// system
#include <assert.h>
#define _USE_MATH_DEFINES
#include <math.h>

// gl stuff
#include <gl/gl.h>
#include "glutils.h"

// physics stuff
#include "shared/physics/2d/2d_body_hull.h"
#include "shared/physics/2d/2d_body_list.h"
#include "shared/physics/2d/2d_body_box.h"
#include "shared/physics/2d/2d_body_sphere.h"
#include "shared/physics/2d/rigid/2d_rigid_sim.h"
#include "shared/physics/2d/rigid/2d_rigid_drag_force.h"
#include "shared/physics/2d/rigid/2d_rigid_gravity_force.h"
#include "shared/physics/2d/rigid/2d_rigid_spring_force.h"
#include "shared/physics/2d/rigid/2d_rigid_control_force.h"
#include "shared/input/joystick.h"
using namespace ajb;
using namespace Physics::D2;
using namespace Physics::D2::Rigid;

// misc
#include "utils.h"
#include <vector>
using namespace std;

// media lib
#include "music.h"
#include "zsound.h"
#include "bitmap.h"

#include "property_system.h"
#include "property_system_editor.h"
#include "editor.h"

#pragma warning(disable:4305; disable: 4244)

/********************************* globals **********************************/

// window/camera stuff
char const AppName[] = "B O O T L O O T E R ! ! !";
int WindowWidth = 800;
int WindowHeight = 600;
bool AppQuit = false;  // stuff this to kill yourself
float FOVY = 45.0f, ZNear = 0.1f, ZFar = 1000.0f;  // whatever, we're 2D
float AspectRatio = float(WindowWidth)/float(WindowHeight); // updated in Resize
float CameraAngleRadians = 0.0, CameraX = 0.0, CameraY = 0.0, CameraZ = 60.0;
Vec2 WorldFromWindowCoords( Vec2 window );
Vec2 WindowFromWorldCoords( Vec2 world );

// win32 stuff
HWND WindowHandle;
HDC WindowDC;

// gl stuff
HGLRC WindowRC;

// functions for you to play in
void Resize( int w, int h );
void Initialize( void );
void Uninitialize( void );
void Idle( void );
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

// system
Music music;

// physics stuff

// needed for game consistency on different machines
#include "../../utils/misc/framerate.h"

Sim simulator;
Drag_Force    global_drag(.10f, .20f);    //a little drag helps with stability
Gravity_Force gravity(Vec2(0.0f, -1.0f));
Rigid_Body    world_body(make_Body_Null(),Vec2(0,0),0,0,0);

char *LevelFileName = "level1.physdb";

PROP_PARSE_THIS_FILE;
PROP_STRUCT struct global_properties
{
    // This is the gravity
    float Gravity;
    
    float MouseSpringKs;
    float MouseSpringKd;
    float MouseSpringRestLength;

    int Foo;
};
static global_properties Globals;
static property *GlobalProperties;
static property_system_editor *GlobalsEditor;

static bool MovingCamera = false;

// game stuff
struct game_object {
    
    game_object(  Rigid_Body *rbInit ) :
            rb(rbInit)
        {
            assert(rb);

            simulator.add(*rb);
        }

    ~game_object( void )
        {
            delete rb;
        }
    
    Rigid_Body *rb;
    float ground_contact; // flattest ground contact y component
    game_object *contact_obj;
    Vec2 contact_point;
    float time_to_death;
    int type;
};
static vector<game_object*> objs;
static vector<Binary_Force *> forces;

static Vec2 window_mouse, world_mouse;
static Joystick PS2Controller;

static Binary_Force *spring = 0;

GLuint
LoadTexture(char *Filename, bool repeat)
{
    GLuint texture;
    
    Bitmap_Load_Result result;
    verify(LoadBitmapByExtension(Filename,&result));

    // gl stuff
    texture = gl_texture_from_bitmap(result.data, result.width, result.height,
                                     result.bytes_per_pixel);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, repeat ? GL_NEAREST_MIPMAP_LINEAR : GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, repeat ? GL_NEAREST_MIPMAP_LINEAR : GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, repeat ? GL_REPEAT : GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, repeat ? GL_REPEAT : GL_CLAMP);

    // delete the bitmap
    delete [] result.data;

    return(texture);
}

enum
{
	T_player,
	T_crate,
    T_game,
    T_boot,
    T_wall,
    T_bullet,
};

static int t_boot, t_wall, t_floor, t_crate, t_title;
static int ellie[6];
static int ellie_n;

game_object *player;

bool face_left;
bool running;
bool jumping;
int kicking;

void
glBody1v(game_object *obj, Body *body, bool Solid)
{
    if(body)
    {
        {
            Body_Hull *p = dynamic_cast<Body_Hull*>(body);
            if(p)
            {
                glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
                for(int i = 0;i<p->num_verts();++i)
                {
                    Vec2 worldspace = p->vert(i);
                    glVertex2f(worldspace.x(),worldspace.y());
                }
                glEnd();
            }
        }

        {
            Body_Box *p = dynamic_cast<Body_Box*>(body);
            if(p)
            {
				float scale_x = 1, scale_y = 1;
                float offset_x = 0, offset_y = 0;
				glEnable(GL_TEXTURE_2D);
				switch (obj->type) {
					case T_boot:
						glBindTexture(GL_TEXTURE_2D, t_boot);
						offset_x = offset_y = 0.5;
						//scale = 4;
						glColor3f(1,1,1);						
						break;
					case T_wall: {
						glBindTexture(GL_TEXTURE_2D, t_wall);
						//scale_y *= 2;
						float bright = 0.6;	
						glColor3f(0.0,1.0 * bright,0.8 * bright);
						break;
					}
					case T_game:
						glBindTexture(GL_TEXTURE_2D, t_floor);
						scale_x *= 0.5;
						glColor3f(1,1,1);						
						break;
					case T_crate:
						glBindTexture(GL_TEXTURE_2D, t_crate);
						scale_y /= 2;
						scale_x /= 2;
						glColor3f(1,1,1);						
						break;
				}
				offset_x = -0.5 * scale_x * p->bs_vert(0).x();
				offset_y = -0.5 * scale_y * p->bs_vert(0).y();
					
                glBegin(Solid ? GL_POLYGON : GL_LINE_LOOP);
                if (obj->type == T_boot)
					offset_x = offset_y = 0.5;
                
                for(int i = 0;i<p->num_verts();++i)
                {
                    Vec2 worldspace = p->vert(i);
                    Vec2 texcoord = 0.5 * p->bs_vert(i);
                    texcoord = Vec2(texcoord.x() * scale_x, texcoord.y() * scale_y);
                    texcoord += Vec2(offset_x, offset_y );
                    glTexCoord2f(texcoord.x(), 1-texcoord.y());
                    glVertex2f(worldspace.x(),worldspace.y());
                }
                glEnd();
                glDisable(GL_TEXTURE_2D);
            }
        }

        {
            Body_Sphere *p = dynamic_cast<Body_Sphere*>(body);
            if(p)
            {
				if (obj->type == T_bullet) return;
				if (obj == player) {
					float w,h;
					++ellie_n;
					glColor3f(1,1,1);
					static int frame[6] = { 2,4,5,3, };
					int ellie_t = 2;
					if (running) ellie_t = frame[(ellie_n >> 3) % 4];
					if (jumping) ellie_t = 5;
					if (kicking > 0) { --kicking; ellie_t = 0; }
					glBindTexture(GL_TEXTURE_2D, ellie[ellie_t]);
					glEnable(GL_TEXTURE_2D);
					Vec2 c = p->x();
					w = 1.5;
					h = 2;
					c += Vec2(0,1.05);
					float x0=0,x1=1;
					if (face_left) x0 = 1, x1 = 0;
					glBegin(GL_POLYGON);
					  glTexCoord2f(x0,0); glVertex2f(c.x()-w, c.y()+h);
					  glTexCoord2f(x0,1); glVertex2f(c.x()-w, c.y()-h);
					  glTexCoord2f(x1,1); glVertex2f(c.x()+w, c.y()-h);
					  glTexCoord2f(x1,0); glVertex2f(c.x()+w, c.y()+h);
					glEnd();
					glDisable(GL_TEXTURE_2D);
					return;
				}
						
                glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
                int total_steps = 48;
                Vec2 center = p->x();
                if(Solid)
                {
					glColor3f(0,0,0);
                    glVertex2f(center.x(), center.y());
                }
                for(int a = 0;a<total_steps;++a)
                {
                    float Angle = 2.0f * 3.14159265358979323846f * ((float)a / (float)(total_steps - 1));
                    float Radius = p->r();
					glColor3f(cos(Angle*8)/2+0.5, sin(Angle*8)/2+0.5,0.5);
                    Angle += obj->rb->O.z();
                    Vec2 arm(Radius*cos(Angle), Radius*sin(Angle));
                    arm += center;
                    glVertex2f(arm.x(), arm.y());
                }
                glEnd();
            }
        }

        Body_List *l = dynamic_cast<Body_List*>(body);
        if(l)
        {
            {for(int Index = 0;
                 Index < l->num_bodies();
                 ++Index)
            {
                glBody1v(obj, &l->body(Index), Solid);
            }}
        }
    }
}

void
glForce1v(Binary_Force *Force)
{
    if(Force)
    {
        glBegin(GL_LINES);
        glVertex2f(Force->m_a.x_worldS().x(), Force->m_a.x_worldS().y());
        glVertex2f(Force->m_b.x_worldS().x(), Force->m_b.x_worldS().y());
        glEnd();
    }
}

void
ClearMouseSpring(void)
{
    if(spring)
    {
        simulator.remove(*spring);
        delete spring;
        spring = 0;
    }
}

void
LoadTestFile(void)
{
    int num_rb=0, num_force=0;
    Rigid_Body **rb_list;
    Binary_Force **force_list;

    loadPhysicsFile(LevelFileName, 0,0, 1.0/16, 1.0/32,
                    &rb_list, &num_rb, &force_list, &num_force, &world_body);

    for (int i=0; i < num_rb; ++i)
    {
		game_object *o = new game_object(rb_list[i]);
        objs.push_back(o);
        o->time_to_death = 500000;
        o->type = T_game;
        rb_list[i]->m_pAppData = o;
        if ((o->rb->m_Pbody->m_collision_plane & 1) == 0) {
			o->type = T_wall;
        }
		std::string temp  = rb_list[i]->name();
		const char *name  = temp.c_str();
		if (strncmp(name, "boot", 4) == 0)
			o->type = T_boot;
		if (strncmp(name, "crate", 5) == 0)
			o->type = T_crate;
		
		char *z = strstr(name, ":v=");
		if (z) {
			float x,y;
			sscanf(z, ":v=%f,%f", &x,&y);
			rb_list[i]->v = Vec2(x,y);
		}
		z = strstr(name, ":w=");
		if (z) {
			float w;
			sscanf(z, ":w=%f",&w);
			rb_list[i]->w = w;
		}
		z = strstr(name, ":g=");
		if (z) {
			float g;
			sscanf(z, ":g=%f",&g);
			rb_list[i]->k_gravity = g;
		}
		z = strstr(name, ":d=");
		if (z) {
			float d;
			sscanf(z, ":d=%f",&d);
			rb_list[i]->k_damp = d;
		}
    }
    simulator.update_bodies_derived_state();

    for (int i=0; i < num_force; ++i)
    {
        forces.push_back(force_list[i]);
        simulator.add(*force_list[i]);
    }
}

game_object *obj_by_name(char *name)
{
	int n = (int) strlen(name);
	for (int i=0; i < objs.size(); ++i)
		if (0 == strnicmp(name, objs[i]->rb->name().c_str(), n))
			return objs[i];
	return NULL;
}

Control_Force *player_control;
Constant_Force *pcon_2;
int which_level=1;
bool switch_level = false;
float switch_time = 2;
int boot_count;

void initLevel(void)
{
	boot_count = 0;
	LoadTestFile();
	player = obj_by_name("player");
	if (player) {
		player_control = new Control_Force(Handle(*player->rb, Vec2(0,0)));
		pcon_2 = new Constant_Force(Handle(*player->rb, Vec2(0,0)));
		simulator.add(*player_control);
		simulator.add(*pcon_2);
		player->rb->k_damp *= 0.5;
		player->rb->k_gravity *= 2;
		player->type = T_player;
	}
}

void
ClearAll(void)
{
    ClearMouseSpring();
    
    for(int i = 0;i < objs.size();++i)
    {
        assert(objs[i]);
        simulator.remove(*objs[i]->rb);
    }

    for(int i = 0;i < forces.size();++i)
    {
        assert(forces[i]);
        simulator.remove(*forces[i]);
    }
    simulator.remove(*player_control);
    simulator.remove(*pcon_2);

    forces.clear();
    objs.clear();
}

/******************************** Initialize ********************************/
/* the window and graphics subsystem are set up but still hidden            */
void Initialize( void )
{
    PS2Controller.init();

    Globals.Gravity = -30.0f;

    Globals.MouseSpringKs = 128.0f;
    Globals.MouseSpringKd = 32.0f;
    Globals.MouseSpringRestLength = 0.0f;

    GlobalProperties = CloneProperty(PROP_LOCAL_STRUCT(global_properties));
    ConvertStructToProperty(PROP_LOCAL_STRUCT(global_properties), &Globals,
                            GlobalProperties);
    GlobalsEditor = OpenPropertySystemEditor("Globals", 850, 50, 400, 400);
    EditPropertiesInPlace(GlobalsEditor, GlobalProperties);
    
    GetTime(); // throw away initial 0.0 so we're seconds-from-startup

    // gl stuff
    glClearColor(0,0,0,0.0f);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    //glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);

    InitializeFrameRate(60.0f);
    
    t_boot = LoadTexture("content/bitmaps/blackbootz-16x32.psd", false);
    t_wall = LoadTexture("content/bitmaps/pillar1.psd", true); // ladder.blue
    t_floor = LoadTexture("content/bitmaps/girder.red.psd", true);
    t_crate = LoadTexture("content/bitmaps/crate.psd", true);
    ellie[0] = LoadTexture("content/bitmaps/elliemay.a.psd", false);
    ellie[1] = LoadTexture("content/bitmaps/elliemay.b.psd", false);
    ellie[2] = LoadTexture("content/bitmaps/elliemay.c.psd", false);
    ellie[3] = LoadTexture("content/bitmaps/elliemay.d.psd", false);
    ellie[4] = LoadTexture("content/bitmaps/elliemay.e.psd", false);
    ellie[5] = LoadTexture("content/bitmaps/elliemay.f.psd", false);
    t_title = LoadTexture("content/bitmaps/title.psd", true);
    
    // init our view
    Resize(WindowWidth,WindowHeight);

    // music init
    music.init();
    music.set_subdirectory("content/music/");
    music.setupnotes();
    music.loadseg("background", "simple_loop.sgt", MUSIC_TRACK_MAIN);

    // sound init
    zsoundInit(WindowHandle);

	zsoundPlay("content/music/shoe_song.wav", -1, 1.0);

    // initialize the physics
    simulator.add(global_drag);
    simulator.add(gravity);
    simulator.add(world_body);

    initLevel();
}

enum
{
	LEFT,
	RIGHT,
	UP,
	DOWN,
	SPACE,
	
	CONTROL__num
};

bool control[CONTROL__num];
bool last[CONTROL__num];

#define MIN_GROUND_ANGLE	0.4
#define MAX_VEL				10
#define CONTROL_KS			2
#define JUMP_SPEED			3
#define SKY_FORCE			0.4
#define JUMP_IMPULSE		3
#define KICK_SPACE			0.2
#define KICK_MASS			3
#define KICK_VEL			22
#define KICK_TIME			0.08

void shoot(void)
{
	Vec2 loc = player->rb->x;
	if (face_left)
		loc -= Vec2(KICK_SPACE,0);
	else
		loc += Vec2(KICK_SPACE,0);
		
	Rigid_Body *r = new Rigid_Body(make_Body_Sphere(0.25), loc, 0,
						1.0 / KICK_MASS, 0);
	game_object *o = new game_object(r); 
	o->type = T_bullet;
	o->time_to_death = KICK_TIME;
	r->m_Pbody->m_collision_plane = ~1;
	if (face_left)
		r->v = Vec2(-KICK_VEL, 0);
	else
		r->v = Vec2( KICK_VEL, 0);
	r->k_gravity = 0;
	
	objs.push_back(o);						
}

void playerSim(float dt)
{
	static float jump_impulse_timer=0;
	static float jump_timer = 0;
	static float jump_contact;
	
	jump_timer -= dt;
	if (jump_timer < 0) jump_timer = 0;

	jump_impulse_timer -= dt;
	if (jump_impulse_timer < 0) jump_impulse_timer = 0;

	float angle = player->ground_contact;
	if (angle < MIN_GROUND_ANGLE)
		angle = jump_contact;
	float weight = (angle - MIN_GROUND_ANGLE) / (1.0 - MIN_GROUND_ANGLE);
	player_control->m_max_v = weight * MAX_VEL;
	player_control->m_Ks    = weight * CONTROL_KS;

	bool full_control = (player->ground_contact > MIN_GROUND_ANGLE || jump_timer > 0);

	bool up = control[UP] && full_control;

	if (full_control) {
		jumping = false;
		if (player->ground_contact > MIN_GROUND_ANGLE && control[UP]) {
			jumping = true;
			jump_timer = 0.1; 
			jump_contact = player->ground_contact;
			if (jump_impulse_timer == 0) {
				player->contact_obj->rb->apply_impulse_at(
					Vec2(0,-JUMP_IMPULSE),
					player->contact_point - player->contact_obj->rb->x
				);
				jump_impulse_timer = 0.1;
			}
		}
			
		if (control[LEFT] != control[RIGHT]) {
			running = true;
			if (control[LEFT]) {
				player_control->m_target = up ? Vec2(-1,JUMP_SPEED) : Vec2(-1,0);
				face_left = true;
			} else {
				player_control->m_target = up ? Vec2( 1,JUMP_SPEED) : Vec2( 1,0);
				face_left = false;
			}
			player_control->m_max_v *= player_control->m_target.len();
			player_control->m_target = player_control->m_target.norm();
		} else {
			running = false;
			if (up) {
				player_control->m_target = Vec2(0,JUMP_SPEED);
			} else {
				player_control->m_target = Vec2(0.5,0);
				player_control->m_max_v  = 0;
			}
		}
		pcon_2->m_f = Vec2(0,0);
	} else {
		running = false;
		jumping = true;
		player_control->m_target = Vec2(0,0);
		if (control[LEFT] != control[RIGHT])
			if (control[LEFT])
				pcon_2->m_f = Vec2(-SKY_FORCE,0);
			else
				pcon_2->m_f = Vec2( SKY_FORCE,0);
		else
			pcon_2->m_f = Vec2(0,0);
	}

	static bool added;
	if (control[SPACE] && !last[SPACE]) {
		shoot();
		kicking = 15;
	}
	memcpy(last, control, sizeof(last));
}

void deleteObj(game_object *o)
{
	simulator.remove(*o->rb);
	objs.erase(std::find(objs.begin(), objs.end(), o));
}

void gameSim(float dt)
{
	int i;
	for (i=0; i < objs.size(); ++i) {
		objs[i]->ground_contact = 0;
		objs[i]->time_to_death -= dt;
	}
    
    for (Contact_List::const_iterator cit=simulator.m_contacts.begin(); cit != simulator.m_contacts.end(); ++cit) 
    {
    //         const Contact& c = *cit;
		game_object *a = (game_object *) (((Rigid_Body *)(cit->m_Prb_a))->m_pAppData);
		game_object *b = (game_object *) (((Rigid_Body *)(cit->m_Prb_b))->m_pAppData);
		if (a != 0 && b != 0) {
			Vec2 n = cit->m_n;
			if (b == player) {
				game_object *t = a;
				a = b;
				b = t;
				n = -n;
			}
			if (a == player) {
				Vec2 loc = cit->m_x;
				// has to be in contact with a point within limited orientation downwards
				Vec2 delta = loc - player->rb->x;
				if (b->type == T_boot) {
					b->time_to_death = -1;
					++boot_count;
                    zsoundPlay("content/sounds/shoe_1.wav");
					if (boot_count >= 2) {
						switch_level = true;
                    }
				} else {
					float ori = atan2(delta.y(), delta.x());
					if (n.y() > a->ground_contact) {
						a->ground_contact = n.y();
						a->contact_obj = b;
						a->contact_point = loc;
					}
				}
			}
	    }
	}
	
	if (player)
		playerSim(dt);

	#define MAX_DELETE	1000		
	static game_object *delete_list[MAX_DELETE];
	int n=0;
	for (i=0; i < objs.size(); ++i)
		if (objs[i]->time_to_death <= 0) {
			delete_list[n++] = objs[i];
			if (n == MAX_DELETE) break;
		}
	for (i=0; i < n; ++i)
		deleteObj(delete_list[i]);
}


void pickLevel(int w);

/*********************************** Idle ***********************************/
/* called by the message loop, and by paint messages                        */
void Idle( void )
{
#if 0
    ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "global_properties"),
                            GlobalProperties,
                            &Globals);
#endif

    gravity.m_g.set_y(Globals.Gravity);
    
    PS2Controller.sample();
    Vec2 LeftStick = PS2Controller.left_stick();
    Vec2 RightStick = PS2Controller.right_stick();

    /**************************** step the simulator ****************************/   

    if (spring)                             //update
        spring->m_a = Handle(world_body, world_mouse);

    int const NumberOfSimSteps = 3;
    for(int StepCounter = 0;StepCounter < NumberOfSimSteps;++StepCounter) {
        // step as many times as you like
        simulator.step(1/(NumberOfSimSteps*GetFrameRate()));
        // do any game stuff you want here (like per-step collision response, etc.)
        gameSim(1/(NumberOfSimSteps*GetFrameRate()));
    }
    // now wait for the rest of the time so we're consistent on fast machines
    WaitFrameRate(GetFrameRate());

    /**************************** render everything *****************************/
   
    // lots of gl stuff
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glDisable(GL_DEPTH_TEST);

    glLoadIdentity();
    glRotated(180.0*(-CameraAngleRadians/M_PI),0,0,1);
    glTranslatef(-CameraX,-CameraY,-CameraZ);

    // draw axes
    //glBegin(GL_LINES);
    //glColor3f(1,0,0);  glVertex2f(0,0);  glVertex2f(1,0);
    //glColor3f(0,1,0);  glVertex2f(0,0);  glVertex2f(0,1);
    //glEnd();

	#if 0
    // draw mouse
    glPointSize(4.0);
    glBegin(GL_POINTS);
    glColor3f(0,0,0);  glVertex2fv(world_mouse.m_v);
    glEnd();
    #endif

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // iterate all the bodies, rendering them
    for(int i = 0;i < objs.size();++i) 
    {
        if (objs[i] == player) continue;
        assert(objs[i]);
        game_object &obj = *objs[i];
        switch (obj.type) {
			case T_game  : glColor3f(0.3,0.3,0.8); break;
			case T_wall  : glColor3f(0.8,0.8,0.6); break;
			case T_player: glColor3f(0,0,0); break;
			case T_bullet: glColor3f(0.5,0.5,0.5); break;
		}
        glBody1v(&obj, obj.rb->m_Pbody.get(), true);
	    glColor3f(0,0,0);
        //glBody1v(&obj, obj.rb->m_Pbody.get(), false);
    }
    
    glColor3f(0.5,0.5,0.5);
    glBody1v(player, player->rb->m_Pbody.get(), true);
    

	#if 0
    // iterate all the forces, rendering them
    glColor3f(1,0,0);
    for(int i = 0;i < forces.size();++i)
    {
        assert(forces[i]);
        Binary_Force *Force = forces[i];

        glForce1v(Force);
    }
    glForce1v(spring); 
    #endif

    glDisable(GL_BLEND);

	#if 0
    glColor3f(0, 0, 1);
    glBegin(GL_LINES);
    float StickScale = 10.0f;
    glVertex2f(0, 0);
    glVertex2f(StickScale * LeftStick.x(), StickScale * LeftStick.y());
    glVertex2f(0, 0);
    glVertex2f(StickScale * RightStick.x(), StickScale * RightStick.y());
    glEnd();
	#endif
	
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	glColor3f(0.5,0.5,1.0);
	char buf2[256];
	sprintf(buf2, "BOOT COUNT: %d", boot_count);
	DrawString(50,50,buf2);
	
#if 0
    glColor3f(0, 1, 0);
    char buf[100];
    sprintf(buf,"wm: %5.1f, %5.1f    sm: %5.1f, %5.1f",
            world_mouse.x(),world_mouse.y(),window_mouse.x(),window_mouse.y());
    DrawString(50,50,buf);

    char *ButtonStrings[] =
    {
        "BUTTON_RIGHT_D",
        "BUTTON_RIGHT_U",
        "BUTTON_RIGHT_L",
        "BUTTON_RIGHT_R",
        "BUTTON_SHOULDER_R1",
        "BUTTON_SHOULDER_R2",
        "BUTTON_SHOULDER_L1",
        "BUTTON_SHOULDER_L2",
        "BUTTON_RSTICK",
        "BUTTON_LSTICK",
        "BUTTON_START",
        "BUTTON_SELECT",
    };

    int Y = 70;
    for(int ButtonIndex = 0;
        ButtonIndex < Joystick::NUM_BUTTONS;
        ++ButtonIndex)
    {
        if(PS2Controller.button_is_down(ButtonIndex))
        {
            DrawString(50,(float) Y,ButtonStrings[ButtonIndex]);
            Y += 10;
        }
    }

    DrawTimingBars();
#endif

	static float fade = 2.2;
	if (fade > 0) {
		float f = fade > 1 ? 1 : fade;
		fade -= (1.0 / GetFrameRate())/2;
		glEnable(GL_BLEND);
		glColor4f(0,0,0,fade);
		glBegin(GL_POLYGON);
		    glVertex2f(-100,-100);
		    glVertex2f( 100,-100);
		    glVertex2f( 100, 100);
		    glVertex2f(-100, 100);
		glEnd();
		
		float g = fade;
		if (g > 1) g = 2-g;
		if (g < 0) g = 0;
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		glColor4f(0.5,0.5,1.0, g);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, t_title);
		glBegin(GL_POLYGON);
		    glTexCoord2f(0,1); glVertex2f(-48,-24);
		    glTexCoord2f(1,1); glVertex2f( 48,-24);
		    glTexCoord2f(1,0); glVertex2f( 48, 24);
		    glTexCoord2f(0,0); glVertex2f(-48, 24);
		glEnd();
		glDisable(GL_BLEND);
	}
    
    SwapBuffers(WindowDC);
    
    if (switch_level) {
		switch_time -= 1.0 / GetFrameRate();
		if (switch_time < 0) {
			if (which_level == 9) exit(0);
			pickLevel(which_level+1);
			switch_time = 2.0;
			switch_level = false;
		}
    }
}

void pickLevel(int w)
{
	static char buf[32];
	sprintf(buf, "level%d.physdb", w);
	which_level = w;
	LevelFileName = buf;
    zsoundPlay("content/sounds/level_up_1.wav");
	
	Globals.Gravity = -30; 
	switch (which_level) {
		case 4: Globals.Gravity = -5; break;	
	}

	ClearAll();
	initLevel();
}

/******************************** AppWndProc ********************************/
/* lame to expose this, but the alternatives all suck too                   */
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    switch (msg)
    {
        // keyboard
        case WM_CHAR:
			switch (wParam)
			{
				case '0': pickLevel(0); break;
				case '1': pickLevel(1); break;
				case '2': pickLevel(2); break;
				case '3': pickLevel(3); break;
				case '4': pickLevel(4); break;
				case '5': pickLevel(5); break;
				case '6': pickLevel(6); break;
				case '7': pickLevel(7); break;
				case '8': pickLevel(8); break;
				case '9': pickLevel(9); break;
			}
            break;

        case WM_MOUSEWHEEL:
        {
            float d = float((short)HIWORD(wParam))/float(WHEEL_DELTA);
            CameraZ *= 1.0f + d/10.0f;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }
        case WM_LBUTTONDOWN:
        {
            SetCapture(hwnd);

            Vec2 body_space_position;
            Rigid_Body* rb = simulator.pick(body_space_position, world_mouse);
            if (rb)
            {
                spring = new Spring_Force(
                    Handle(world_body, world_mouse),
                    Handle(*rb, body_space_position), 
                    Globals.MouseSpringKs,
                    Globals.MouseSpringKd,
                    Globals.MouseSpringRestLength
                    );
                simulator.add(*spring);
            }
        } break;

        case WM_LBUTTONUP:
        {
            if(!(wParam & MK_MBUTTON & MK_RBUTTON))
            {
                ReleaseCapture();
            }

            ClearMouseSpring();
        } break;

        case WM_RBUTTONDOWN:
        {
            SetCapture(hwnd);

            if(spring)
            {
                forces.push_back(spring);
                spring = 0;
            }
            break;
        }
        case WM_MBUTTONDOWN:
        {
            MovingCamera = true;
        } break;
        case WM_MBUTTONUP:
        {
            MovingCamera = false;
        } break;

        case WM_RBUTTONUP:
        {
            if(!(wParam & MK_MBUTTON & MK_LBUTTON))
            {
                ReleaseCapture();
            }
        } break;
            
        case WM_KEYDOWN: {
            switch(wParam) {
				
                case VK_LEFT: {
                    control[LEFT] = true;
                    break;
                }
                case VK_RIGHT: {
                    control[RIGHT] = true;
                    break;
                }
                case VK_UP: {
					control[UP] = true;
					break;
				}
				case VK_DOWN: {
					control[DOWN] = true;
					break;
				}
				case ' ': {
					control[SPACE] = true;
					break;
				}

                case 'M': {
                    // demo the audio stuff
                    if(0x8000 & GetKeyState(VK_SHIFT)) {
                        music.play("background");
                    } else {
                        zsoundPlay("content/sounds/bad.0002.wav");
                    }
                    break;
                }

                case 'E': {
                    launchEditor(LevelFileName);
                    ClearAll();
                    initLevel();
                } break;
                
                case 'R': {
					ClearAll();
					initLevel();
					break;
                }
            
                case VK_ESCAPE: {
                    AppQuit = true;
                    break;
                }
            }
            break;
        }
        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
            break;
        case WM_KEYUP:
            switch(wParam) {
                case VK_LEFT: {
                    control[LEFT] = false;
                    break;
                }
                case VK_RIGHT: {
                    control[RIGHT] = false;
                    break;
                }
                case VK_UP: {
					control[UP] = false;
					break;
				}
				case VK_DOWN: {
					control[DOWN] = false;
					break;
				}
				case ' ': {
					control[SPACE] = false;
					break;
				}
			}
            break;

        case WM_MOUSEMOVE: {
            Vec2 new_window_mouse = Vec2(float(GET_X_LPARAM(lParam)),float(GET_Y_LPARAM(lParam)));
            if(MovingCamera)
            {
                Vec2 d_world = WorldFromWindowCoords(new_window_mouse) - world_mouse;
                CameraX -= d_world.x();
                CameraY -= d_world.y();
            }
            window_mouse = new_window_mouse;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }

         
            /**************** below is stuff you probably don't care about ****************/
         
        case WM_DESTROY: {
            PostQuitMessage(0);
            break;
        }
        case WM_PAINT: {
            PAINTSTRUCT ps;
            BeginPaint(hwnd,&ps);
            Idle();
            EndPaint(hwnd,&ps);
            return 0L;
        }
        case WM_SIZE: {
            Resize(LOWORD(lParam),HIWORD(lParam));
            break;
        }
    }
    return (LONG)DefWindowProc(hwnd,msg,(LONG)wParam,(LONG)lParam);
}

/********************************** Resize **********************************/
/* called on WM_SIZE, and at the bottom of initialize                       */
void Resize( int w, int h )
{
    // compute our frustrum planes, the x,y plane values are on the near clip plane
    WindowWidth = w;
    WindowHeight = h;
    AspectRatio = float(w)/float(h);
    float ymax = ZNear * float(tan(FOVY/2.0));
    float ymin = -ymax;
    float xmin = AspectRatio * ymin;
    float xmax = AspectRatio * ymax;
   
    // load the viewport and projection matrix, gl stuff
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // gluPerspective doesn't set the near plane as the projection
    // plane, so you can't do good math, so do it ourselves
    glFrustum(xmin,xmax,ymin,ymax,ZNear,ZFar);
    glMatrixMode(GL_MODELVIEW);
}

/******************************* Uninitialize *******************************/
/* called before the graphics is destroyed, but the window is already toast */
void Uninitialize( void )
{
}

/********************************* WinMain **********************************/

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
    // register and create our window
    WNDCLASS cls;
    if(!hPrev) {
        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = 0;
        cls.lpszMenuName   = 0;
        cls.lpszClassName  = AppName;
        cls.hbrBackground  = 0;
        cls.hInstance      = hInst;
        cls.style          = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS | CS_OWNDC;
        cls.lpfnWndProc    = (WNDPROC)AppWndProc;
        cls.cbWndExtra     = 0;
        cls.cbClsExtra     = 0;

        if (!RegisterClass(&cls))
            return FALSE;
    }

    DWORD Style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    RECT WindowRect = { 0, 0 };
    WindowRect.right = WindowWidth;
    WindowRect.bottom = WindowHeight;
    BOOL Menu = FALSE;

    AdjustWindowRect(&WindowRect,Style,Menu);
    
    WindowHandle = CreateWindow(AppName,AppName,Style,
                                0,50,
                                WindowRect.right-WindowRect.left,
                                WindowRect.bottom-WindowRect.top,
                                0,0,hInst,0);

    // create our gl stuff
    WindowDC = GetDC(WindowHandle);
    PIXELFORMATDESCRIPTOR FormatDescriptor = {
        sizeof (PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA, 32,   // 32bpp
        0,0, 0,0, 0,0, 0,0,
        0, 0,0,0,0,
        24, 8,               // depth, stencil
        0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };
    int FormatIndex = ChoosePixelFormat(WindowDC,&FormatDescriptor);
    SetPixelFormat(WindowDC,FormatIndex,&FormatDescriptor);
    WindowRC = wglCreateContext(WindowDC);
    wglMakeCurrent(WindowDC,WindowRC);

    // okay, ready to go
   
    Initialize();    // call the app
   
    ShowWindow(WindowHandle,sw);
   
    // pump the message queue
    MSG msg;
    while(!AppQuit) {
        if(PeekMessage(&msg,0,0,0,PM_REMOVE)) {
            if(msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
            Idle();    // call the app
        }
    }

    Uninitialize();  // call the app
   
    // clean up gl stuff and exit
    wglMakeCurrent(0,0);
    wglDeleteContext(WindowRC);
    ReleaseDC(WindowHandle,WindowDC);
   
    return int(msg.wParam);
}

/********************** utilities that use our globals **********************/
Vec2 WorldFromWindowCoords( Vec2 window )
{
    float xn = float(window.x())/WindowWidth - 0.5f;
    float yn = float(WindowHeight-window.y())/WindowHeight - 0.5f;
    float xc = float(xn*2.0*CameraZ*tan(FOVY/2)*AspectRatio);
    float yc = float(yn*2.0*CameraZ*tan(FOVY/2));
    return Vec2(CameraX,CameraY) + Matrix22::rotation(CameraAngleRadians) * Vec2(xc,yc);
}
Vec2 WindowFromWorldCoords( Vec2 world )
{
    Vec2 c = Matrix22::rotation(-CameraAngleRadians) * (world - Vec2(CameraX,CameraY));
    float xn = float(c.x()/(2.0*CameraZ*tan(FOVY/2)*AspectRatio));
    float yn = float(c.y()/(2.0*CameraZ*tan(FOVY/2)));
    float xs = WindowWidth*(xn + 0.5f);
    float ys = WindowHeight - WindowHeight*(yn + 0.5f);
    return Vec2(xs,ys);
}
